# Deployment

## Local Deployment

### System Requirements
Before you begin installing the Polaris Learning Infrastructure, ensure that your system meets the following requirements:

#### Recommended System Requirements:
- **Operating System**: Linux (Debian 11+, Ubuntu 20.04+ recommended)
- **Processor**: At least 4 CPU cores (8 recommended for larger installations)
- **RAM**: Minimum of 8 GB RAM (16 GB or more recommended for larger user bases)
- **Storage**: Minimum of 50 GB free disk space (expandable based on data volume)
- **Network**: Stable internet connection for updates and container downloads

#### Software Requirements:
- **Docker** (minimum version 20.10)
- **Docker Compose** (minimum version 1.29)
- **Git** (for downloading configuration files)
- **cURL or Wget** (for fetching external dependencies)

### Installing Docker and Docker Compose
If Docker is not yet installed, you can set it up on Debian or Ubuntu using the following commands:

```bash
# Add Docker's official GPG key:
sudo apt-get update
sudo apt-get install ca-certificates curl
sudo install -m 0755 -d /etc/apt/keyrings
sudo curl -fsSL https://download.docker.com/linux/ubuntu/gpg -o /etc/apt/keyrings/docker.asc
sudo chmod a+r /etc/apt/keyrings/docker.asc

# Add the repository to Apt sources:
echo \
  "deb [arch=$(dpkg --print-architecture) signed-by=/etc/apt/keyrings/docker.asc] https://download.docker.com/linux/ubuntu \
  $(. /etc/os-release && echo "${UBUNTU_CODENAME:-$VERSION_CODENAME}") stable" | \
  sudo tee /etc/apt/sources.list.d/docker.list > /dev/null
sudo apt-get update
```

Install Docker and Docker Compose:

```bash
sudo apt-get install docker-ce docker-ce-cli containerd.io docker-buildx-plugin docker-compose-plugin
```

Verify that Docker is installed:

```bash
docker --version
docker-compose --version
sudo docker run hello-world
```

### Docker Login to the RWTH Aachen Registry
If you are not already logged in, you must authenticate using your GitLab username and password.
If you have enabled two-factor authentication (2FA), you will need a Personal Access Token (PAT) instead.

```bash
echo "YOUR-TOKEN" | docker login registry.git.rwth-aachen.de -u YOUR-USERNAME --password-stdin
```

After a successful login, the following message will appear:

```bash
Login Succeeded
```

#### Create Private/Public Keys and .env File

Please change all passwords in the configuration file and leave the passphrase empty.

```bash
ssh-keygen -b 4096 -f id_rsa
cp .env.sample .env
```

#### Start Containers

```bash
docker compose up -d
```

Please check whether all services started successfully:

```bash
docker compose ps
```

After that, you should be able to visit http://localhost:8004/ and see the rights engine.


## Expermimental: Remote Deployment via Gitlab CI Pipeline

It is recommended to deploy the rights engine in production through a Gitlab CI pipeline. The process is currrently expermimental and need an adjustment to the concrete setup. 

### Remote Deployment on three VMs

Instead of using a Gitlab CI Pipeline, one can use three VMs for the Polaris deployment. Each VM is responsible one of the three Polaris components _Rights Engine_, _Analytics Engine_ and _Mongo (LRS)_.

First, for each VM, the [deployment repository](https://git.rwth-aachen.de/polaris/entwicklung/deployment) is checked out, configured, and the appropriate Docker services are started.

> Care should be taken to always check out the branch `digitallearning`.

Requirements 1. Machines must be able to access the Gitlab from Aachen, full internet access would be better 2. docker and docker compose must be installed 3. SSL certificates and domains must be available 4. Machines must be able to communicate with each other via HTTP and HTTPS

> Passwords should not contain special characters, since this often lead to problems. 

> Docker creates folder, if the corresponding volume not exists. 

#### Mongo (LRS) VM setup

##### 1. Clone deployment repository
Clone the deployment repository for a quick start. 

```console
$ git clone https://git.rwth-aachen.de/polaris/entwicklung/deployment.git
```

#### 2. Configuration
Next, we update the .env file of the Mongo database. 

Settings for `MONGO_USER` and `MONGO_PASSWORT` need to match settings in `.env` for Rights Engine.

```console
$ cd mongodb
$ cp .env.sample .env
$ vi .env
```

#### 3. Create Docker network
Next, we need to create a Docker network that enables a secure network connection between the mongo database and the rights engine of Polaris.

```console
$ docker network create web
```

If you like to check if the network is created, use the following command

```console
$ docker network ls
```

#### 4. Starting Docker services
Finally, start the Mongo database with the following command

```console
$ docker compose up -d
```

Check, if the Mongo database is running:

```console
$ docker compose ps
```

### Rights Engine VM setup

#### 1. Clone deployment repository
Clone the deployment repository for a quick start. 

```console
$ git clone https://git.rwth-aachen.de/polaris/entwicklung/deployment.git
```

#### 2. Configuration

```console
$ cd rights-engine
$ cp .env.sample .env
$ vi .env
```

#### 3. Generate key pair

```console
$ ssh-keygen -b 4096 -f id_rsa
```

#### 4. Create Docker network

```console
$ docker network create web
```

#### 5. Install Traefik

See section `Installing Traefik`

#### 6. Starting Docker services

```console
$ docker compose up -d
```

### Analytics Engine VM setup

#### 1. Clone deployment repository

```console
$ git clone https://git.rwth-aachen.de/polaris/entwicklung/deployment.git
```

#### 2. Checkout `digitallearning` branch

```console
$ git checkout digitallearning
```

#### 3. Configuration

```console
$ cd analytics-engine
$ cp .env.sample .env
$ vi .env
```

#### 4. Create Docker network

```console
$ docker network create web
```

#### 5. Install Traefik

See section `Installing Traefik`

#### 6. Starting Docker services

```console
$ docker compose up -d
```

#### 7. Migrate DB

```console
$ docker compose exec -it scheduler sh -c 'scheduler create-db'
```

## Update Docker Images

```bash
docker compose pull
```

## Installing Traefik

```console
$ mkdir traefik && cd traefik
```

```console
$ touch docker-compose.yml
```

Content `docker-compose.yml`

```yml
services:
  traefik:
    volumes:
      - "/var/run/docker.sock:/var/run/docker.sock:ro"
      - "$PWD/traefik.toml:/traefik.toml"
      - "$PWD/traefik_dynamic.toml:/traefik_dynamic.toml"
      - "$PWD/acme.json:/acme.json"
      - "$PWD/access.log:/access.log"
    ports:
      - "80:80"
      - "443:443"
      - "8082:8080"
    network_mode: web
    container_name: traefik
    restart: always
    image: "traefik:v2.9"
```

```console
$ touch acme.json && touch access.log && chmod 600 acme.json
$ touch traefik_dynamic.toml
```

Content `traefik_dynamic.toml`

```toml
[http.middlewares.simpleAuth.basicAuth]
  users = [
    "admin:$apr1$SaCzhK8I$aShLLTn2wYgq3MyOQyTrG."
  ]

[http.routers.api]
  rule = "Host(`polaris.uni-bochum.de`)"
  entrypoints = ["websecure"]
  middlewares = ["simpleAuth"]
  service = "api@internal"
  [http.routers.api.tls]
    certResolver = "lets-encrypt"

[http.middlewares]
    [http.middlewares.security-headers.headers]
      # CORS
      #AccessControlAllowMethods = ["GET", "OPTIONS", "PUT"]
      #AccessControlAllowOrigin = "origin-list-or-null"
      #AccessControlMaxAge = 100
      #AddVaryHeader = true
      BrowserXssFilter = true
      ContentTypeNosniff = true
      ForceSTSHeader = true
      FrameDeny = true
      SSLRedirect = true
      STSIncludeSubdomains = true
      STSPreload = true
      #ContentSecurityPolicy = "default-src 'self' 'unsafe-inline'"
      CustomFrameOptionsValue = "SAMEORIGIN"
      #ReferrerPolicy = "same-origin"
      #FeaturePolicy = "vibrate 'self'"
      STSSeconds = 315360000
[tls.options]
  [tls.options.default]
    minVersion = "VersionTLS12"
    curvePreferences = ["CurveP521", "CurveP384"]
    sniStrict = true
    cipherSuites = [
      "TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256",
      "TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384",
      "TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305",
      "TLS_AES_128_GCM_SHA256",
      "TLS_AES_256_GCM_SHA384",
      "TLS_CHACHA20_POLY1305_SHA256",
    ]
```

```console
$ touch traefik.toml
```

Content `traefik.toml`

- Please change __email@email.de__

```toml
[entryPoints]
  [entryPoints.web]
    address = ":80"
    [entryPoints.web.http.redirections.entryPoint]
      to = "websecure"
      scheme = "https"

  [entryPoints.websecure]
    address = ":443"

[api]
  dashboard = true

[certificatesResolvers.lets-encrypt.acme]
  email = "email@email.de"
  storage = "acme.json"
  [certificatesResolvers.lets-encrypt.acme.httpChallenge]
    entryPoint = "web"

[providers.docker]
  watch = true
  network = "web"
  exposedByDefault= "false"

[providers.file]
  filename = "traefik_dynamic.toml"

[accessLog]
  filePath = "access.log"
  format = "json"

  [accessLog.filters]
    statusCodes = ["200", "300-302"]
```

## Adjust deployment 

### Number of gunicorn workers 
The number of Gunicorn workers can be adjusted using the WORKERS environment variable.
By default, the application starts with 4 workers. You can override this value during deployment:

```docker-compose.yml
services:
  web:
    image: registry.git.rwth-aachen.de/polaris/rights-engine/rights-engine:${POLARIS_VERSION}
    ports:
      - "80:80"
    environment:
      - WORKERS=10
```