import json
import os
from io import StringIO

from rest_framework.test import APIClient
from rolepermissions.roles import assign_role
from django.core.management import call_command

from consents.tests.tests_consent_operations import BaseTestCase
from providers.models import ProviderAuthorization, Provider, ProviderVerbGroup, ProviderSchema
from users.models import CustomUser

from consents.models import UserConsents


class TestUserEndpoints(BaseTestCase):
    verb_groups = [
        {
            "provider_id": 1,
            "id": "default_group",
            "label": "Default group",
            "description": "default",
            "showVerbDetails": True,
            "purposeOfCollection": "Lorem Ipsum",
            "requiresConsent": True,
            "verbs": [
                {"id": "http://h5p.example.com/expapi/verbs/experienced"},
                {"id": "http://h5p.example.com/expapi/verbs/attempted"},
                {"id": "http://h5p.example.com/expapi/verbs/interacted"},
                {"id": "http://h5p.example.com/expapi/verbs/answered"}
            ]
        },
    ]

    def setUp(self):
        call_command('check_and_apply_migrations')
        normal_user = CustomUser.objects.create_user(
            self.test_user_email, self.test_user_password
        )
        provider_user = CustomUser.objects.create_user(
            self.test_provider_email, self.test_provider_password
        )

        assign_role(normal_user, "user")
        assign_role(provider_user, "provider_manager")
        assign_role(provider_user, 'analyst') # needed to create analytics tokens later on

        response = self.client.post(
            "/api/v1/auth/token",
            {"email": self.test_user_email, "password": self.test_user_password},
        )  # obtain token
        self.assertEqual(response.status_code, 200)
        self.assertTrue("access" in response.data)
        user_token = response.data["access"]

        response = self.client.post(
            "/api/v1/auth/token",
            {
                "email": self.test_provider_email,
                "password": self.test_provider_password,
            },
        )  # obtain token
        self.assertEqual(response.status_code, 200)
        self.assertTrue("access" in response.data)
        provider_token = response.data["access"]

        self.user_client = APIClient()
        self.user_client.credentials(HTTP_AUTHORIZATION="Bearer " + user_token)

        self.provider_client = APIClient()
        self.provider_client.credentials(HTTP_AUTHORIZATION="Bearer " + provider_token)

        # Create provider schema
        with StringIO(json.dumps(self.h5p_provider_schema)) as fp:
            response = self.provider_client.put(
                "/api/v1/consents/provider/create",
                {"provider-schema": fp},
                format="multipart",
            )
            self.assertEqual(response.status_code, 201)

        # Get application token for created provider
        self.application_token = ProviderAuthorization.objects.get(
            provider__name="H5P"
        ).key
        self.third_party_client = APIClient()
        self.third_party_client.credentials(
            HTTP_AUTHORIZATION="Basic " + self.application_token
        )

        # create some verb groups
        for group in self.verb_groups:
            response = self.provider_client.post(
                "/api/v1/consents/provider/" + str(Provider.objects.latest('id').id) + "/create-verb-group",
                group,
                format="json",
            )
            self.assertEqual(response.status_code, 200)

    def test_get_consent_for_provider(self):
        # Note: this route is currently not being used in the frontend since the consent history replaces most of the functionality
        provider_id = Provider.objects.latest('id').id

        response = self.user_client.get(
            f"/api/v1/consents/user/{provider_id}",
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())

        # right now we haven't declared consent for any groups
        self.assertEqual(data["consent"], None)

        # create consent
        group_id = ProviderVerbGroup.objects.latest('id').id # maybe todo: rework frontend to use the "group_id" attribute instead of primary keys
        response = self.user_client.post(
            "/api/v1/consents/user/save", {"groups": [group_id]}, format="json"
        )
        self.assertEqual(response.status_code, 200)

        # now we would expect the consent to be present in the response
        response = self.user_client.get(
            f"/api/v1/consents/user/{provider_id}",
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())

        self.assertTrue(all(key in data.keys() for key in ["consent", "provider_schema", "paused_data_recording"]))
        self.assertFalse(data["consent"] is None)

    def test_get_consent_history(self):
        provider_id = Provider.objects.latest('id').id

        # consent history should be empty
        response = self.user_client.get(
            "/api/v1/consents/user/history",
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())
        self.assertTrue("groups" in data.keys())
        self.assertEqual(len(data["groups"]), 0)

        # create consent
        group_id = ProviderVerbGroup.objects.latest('id').id
        response = self.user_client.post(
            "/api/v1/consents/user/save", {"groups": [group_id]}, format="json"
        )
        self.assertEqual(response.status_code, 200)

        response = self.user_client.get(
            "/api/v1/consents/user/history",
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())
        self.assertTrue("groups" in data.keys())
        self.assertEqual(len(data["groups"]), 1)

        # create another group
        group = {"id": "default_group", "label": "Group 2",
                 "description": "Lorem ipsum", "showVerbDetails": True, "purposeOfCollection": "Lorem Ipsum",
                 "requiresConsent": True, "verbs": [
                {"id": "http://h5p.example.com/expapi/verbs/experienced"},
                {"id": "http://h5p.example.com/expapi/verbs/interacted"},
                {"id": "http://h5p.example.com/expapi/verbs/answered"},
            ]}
        response = self.provider_client.post(
            f"/api/v1/consents/provider/{provider_id}/create-verb-group",
            group,
            format="json",
        )
        self.assertEqual(response.status_code, 200)

        # create consent for second group
        group_id = ProviderVerbGroup.objects.latest('id').id
        response = self.user_client.post(
            "/api/v1/consents/user/save", {"groups": [group_id]}, format="json"
        )
        self.assertEqual(response.status_code, 200)

        # check that both consents are present in history
        response = self.user_client.get(
            "/api/v1/consents/user/history",
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())
        self.assertTrue("groups" in data.keys())
        self.assertEqual(len(data["groups"]), 2)

    def test_update_consent_group_active(self):
        provider_id = Provider.objects.latest('id').id

        response = self.user_client.get(
            f"/api/v1/consents/user/{provider_id}",
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())

        # right now we haven't declared consent for any groups
        self.assertEqual(data["consent"], None)

        # create consent
        group_id = ProviderVerbGroup.objects.latest('id').id
        response = self.user_client.post(
            "/api/v1/consents/user/save", {"groups": [group_id]}, format="json"
        )
        self.assertEqual(response.status_code, 200)

        # check if incorrect request is rejected
        response = self.user_client.post(
            "/api/v1/consents/user/update-consent-group-active", {}, format="json"
        )
        self.assertEqual(response.status_code, 400)

        # update group consent
        response = self.user_client.post(
            "/api/v1/consents/user/update-consent-group-active", {"id": group_id, "active": False}, format="json"
        )
        self.assertEqual(response.status_code, 200)
        # check if status correctly stored in DB
        self.assertFalse(UserConsents.objects.last().active)

    def test_revoke_consent_group(self):
        provider_id = Provider.objects.latest('id').id

        response = self.user_client.get(
            f"/api/v1/consents/user/{provider_id}",
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())

        # right now we haven't declared consent for any groups
        self.assertEqual(data["consent"], None)

        # create consent
        group_id = ProviderVerbGroup.objects.latest('id').id
        response = self.user_client.post(
            "/api/v1/consents/user/save", {"groups": [group_id]}, format="json"
        )
        self.assertEqual(response.status_code, 200)

        # check if incorrect request is rejected
        response = self.user_client.post(
            "/api/v1/consents/user/revoke-consent-group", {}, format="json"
        )
        self.assertEqual(response.status_code, 400)

        # update group consent
        response = self.user_client.post(
            "/api/v1/consents/user/revoke-consent-group", {"id": group_id}, format="json"
        )
        self.assertEqual(response.status_code, 200)
        # check if consent correctly updated in DB
        self.assertFalse(UserConsents.objects.last().consented)

    def test_get_analytics_tokens(self):
        response = self.provider_client.get(
            f"/api/v1/consents/user/analytics-tokens",
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())
        self.assertEqual(len(data), 0)

        response = self.provider_client.post(
            f"/api/v1/provider/analytics-tokens/create",
            {
                "name": "Test analytics token",
                "description": "Test analytics token",
                "expires": "12/31/2099",
                "can_access": []
            }
        )
        self.assertEqual(response.status_code, 201)

        response = self.provider_client.get(
            f"/api/v1/consents/user/analytics-tokens",
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())
        self.assertEqual(len(data), 1)

    def test_get_analytics_token_consent(self):
        response = self.provider_client.get(
            f"/api/v1/consents/user/analytics-tokens",
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())
        self.assertEqual(len(data), 0)

        response = self.provider_client.post(
            f"/api/v1/provider/analytics-tokens/create",
            {
                "name": "Test analytics token",
                "description": "Test analytics token",
                "expires": "12/31/2099",
                "can_access": []
            }
        )
        self.assertEqual(response.status_code, 201)
        data = json.loads(response.content.decode())
        analyticTokenId = data["id"]

        response = self.user_client.post(
            f"/api/v1/consents/user/analytics-tokens/consent",
            {
                "analyticTokenId": analyticTokenId
            }
        )
        self.assertEqual(response.status_code, 200)
        # maybe todo: check if consents created correctly (?)

    def test_get_providers(self):
        response = self.user_client.get(
            f"/api/v1/consents/user/providers",
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())
        self.assertEqual(len(data), 1)


    def test_get_status(self):
        response = self.user_client.get(
            f"/api/v1/consents/user/status",
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())
        self.assertTrue("result" in data.keys())
        self.assertTrue("1" in data["result"].keys())
        self.assertEqual(data["result"]["1"]["status"], "none") # since consent is coupled with groups, consent for a whole provider schema can not exist