import json
import os
from io import StringIO

from django.conf import settings
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from rest_framework.test import APIClient
from rolepermissions.roles import assign_role
from django.core.management import call_command

from consents.tests.tests_consent_operations import BaseTestCase
from providers.models import ProviderAuthorization, Provider, ProviderVerbGroup, ProviderSchema
from users.models import CustomUser

from data_removal.models import DataRemovalJob


class TestsDataRemoval(BaseTestCase):
    verb_groups = [
        {
            "provider_id": 1,
            "id": "default_group",
            "label": "Default group",
            "description": "default",
            "showVerbDetails": True,
            "purposeOfCollection": "Lorem Ipsum",
            "requiresConsent": True,
            "verbs": [
                {"id": "http://h5p.example.com/expapi/verbs/experienced"},
                {"id": "http://h5p.example.com/expapi/verbs/attempted"},
                {"id": "http://h5p.example.com/expapi/verbs/interacted"},
                {"id": "http://h5p.example.com/expapi/verbs/answered"}
            ]
        },
    ]

    def setUp(self):
        call_command('check_and_apply_migrations')
        normal_user = CustomUser.objects.create_user(
            self.test_user_email, self.test_user_password
        )
        provider_user = CustomUser.objects.create_user(
            self.test_provider_email, self.test_provider_password
        )

        assign_role(normal_user, "user")
        assign_role(provider_user, "provider_manager")

        response = self.client.post(
            "/api/v1/auth/token",
            {"email": self.test_user_email, "password": self.test_user_password},
        )  # obtain token
        self.assertEqual(response.status_code, 200)
        self.assertTrue("access" in response.data)
        user_token = response.data["access"]

        response = self.client.post(
            "/api/v1/auth/token",
            {
                "email": self.test_provider_email,
                "password": self.test_provider_password,
            },
        )  # obtain token
        self.assertEqual(response.status_code, 200)
        self.assertTrue("access" in response.data)
        provider_token = response.data["access"]

        self.user_client = APIClient()
        self.user_client.credentials(HTTP_AUTHORIZATION="Bearer " + user_token)

        self.provider_client = APIClient()
        self.provider_client.credentials(HTTP_AUTHORIZATION="Bearer " + provider_token)

        # Create provider schema
        with StringIO(json.dumps(self.h5p_provider_schema)) as fp:
            response = self.provider_client.put(
                "/api/v1/consents/provider/create",
                {"provider-schema": fp},
                format="multipart",
            )
            self.assertEqual(response.status_code, 201)

        # Get application token for created provider
        self.application_token = ProviderAuthorization.objects.get(
            provider__name="H5P"
        ).key
        self.third_party_client = APIClient()
        self.third_party_client.credentials(
            HTTP_AUTHORIZATION="Basic " + self.application_token
        )

        # create some verb groups
        for group in self.verb_groups:
            response = self.provider_client.post(
                "/api/v1/consents/provider/" + str(Provider.objects.latest('id').id) + "/create-verb-group",
                group,
                format="json",
            )
            self.assertEqual(response.status_code, 200)


    def test_create_job(self):
        self.assertEqual(DataRemovalJob.objects.count(), 0)

        response = self.user_client.post("/api/v1/data-removal/create", {
            "scope": {},
            "immediately": True,
        }, format="json")
        self.assertEqual(response.status_code, 200)

        self.assertEqual(DataRemovalJob.objects.count(), 1)


    def test_list_jobs(self):
        response = self.user_client.get("/api/v1/data-removal/list", )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())
        self.assertEqual(len(data), 0)

        response = self.user_client.post("/api/v1/data-removal/create", {
            "scope": {},
            "immediately": True,
        }, format="json")
        self.assertEqual(response.status_code, 200)

        response = self.user_client.get("/api/v1/data-removal/list",)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())
        self.assertEqual(len(data), 1)