import json
import os

from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import TestCase
from rest_framework.test import APIClient
from rolepermissions.roles import assign_role
from django.core.management import call_command

from users.models import CustomUser

from providers.models import Provider, ProviderAuthorization, AnalyticsToken

PROJECT_PATH = os.path.abspath(os.path.dirname(__name__))

class ProviderTestCase(TestCase):

    test_user_email = "test@mail.com"
    test_user_password = "test123"

    test_provider_email = "test2@mail.com"
    test_provider_password = "test123"

    def setUp(self):
        call_command('check_and_apply_migrations')
        normal_user = CustomUser.objects.create_user(self.test_user_email, self.test_user_password)
        provider_user = CustomUser.objects.create_user(self.test_provider_email, self.test_provider_password)

        assign_role(normal_user, 'user')
        assign_role(provider_user, 'provider_manager')
        assign_role(provider_user, 'analyst')  # needed to create analytics tokens

        response=self.client.post('/api/v1/auth/token', {"email": self.test_user_email, "password": self.test_user_password}) # obtain token
        self.assertEqual(response.status_code, 200)
        self.assertTrue('access' in response.data)
        user_token = response.data['access']

        response=self.client.post('/api/v1/auth/token', {"email": self.test_provider_email, "password": self.test_provider_password}) # obtain token
        self.assertEqual(response.status_code, 200)
        self.assertTrue('access' in response.data)
        provider_token = response.data['access']

        self.user_client = APIClient()
        self.user_client.credentials(HTTP_AUTHORIZATION='Bearer ' + user_token)

        self.provider_client = APIClient()
        self.provider_client.credentials(HTTP_AUTHORIZATION='Bearer ' + provider_token)

        # we need provider schemas for many of these tests
        with open(os.path.join(PROJECT_PATH, "static/provider_schema_h5p_v1.example.json")) as fp:
            response=self.provider_client.put('/api/v1/consents/provider/create', {
                    "provider-schema": fp
                }, format='multipart')
            self.assertEqual(response.status_code, 201)
            self.assertTrue('message' in response.json())

    def test_create_token(self):
        response = self.provider_client.post(
            f"/api/v1/provider/analytics-tokens/create",
            {
                "name": "Test analytics token",
                "description": "Test analytics token",
                "expires": "12/31/2099",
                "can_access": []
            }, format="json"
        )
        self.assertEqual(response.status_code, 201)


    def test_token_name_available(self):
        response = self.provider_client.post(
            f"/api/v1/provider/analytics-tokens/create",
            {
                "name": "Test analytics token",
                "description": "Test analytics token",
                "expires": "12/31/2099",
                "can_access": []
            }, format="json"
        )
        self.assertEqual(response.status_code, 201)

        # test unavailable
        response = self.provider_client.post(
            f"/api/v1/provider/analytics-tokens/name-available",
            {
                "name": "Test analytics token"
            }, format="json"
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())
        self.assertFalse(data["is_available"])

        # test available
        response = self.provider_client.post(
            f"/api/v1/provider/analytics-tokens/name-available",
            {
                "name": "Test analytics token other"
            }, format="json"
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode())
        self.assertTrue(data["is_available"])


        # test failure case
        response = self.provider_client.post(
            f"/api/v1/provider/analytics-tokens/name-available",
            {
                "name": None
            }, format="json"
        )
        self.assertEqual(response.status_code, 400)

    def test_delete_token(self):
        response = self.provider_client.post(
            f"/api/v1/provider/analytics-tokens/create",
            {
                "name": "Test analytics token",
                "description": "Test analytics token",
                "expires": "12/31/2099",
                "can_access": []
            }, format="json"
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(AnalyticsToken.objects.count(), 1)

        token_id = AnalyticsToken.objects.latest('id').id

        response = self.provider_client.delete(f"/api/v1/provider/analytics-tokens/{token_id}/delete")
        self.assertEqual(response.status_code, 204)
        self.assertEqual(AnalyticsToken.objects.count(), 0)

    def test_update_token_verbs(self):
        provider_id = Provider.objects.latest('id').id

        response = self.provider_client.post(
            f"/api/v1/provider/analytics-tokens/create",
            {
                "name": "Test analytics token",
                "description": "Test analytics token",
                "expires": "12/31/2099",
                "can_access": []
            }, format="json"
        )
        self.assertEqual(response.status_code, 201)

        token_id = AnalyticsToken.objects.latest('id').id

        # test wrong token ID
        response = self.provider_client.post(f"/api/v1/provider/analytics-tokens/1000000/update-verbs",
         {
             "active_verbs": [
                 {"verb": "http://h5p.example.com/expapi/verbs/experienced",
                  "provider": provider_id},
                 {"verb": "http://h5p.example.com/expapi/verbs/attempted",
                  "provider": provider_id},
                 {"verb": "http://h5p.example.com/expapi/verbs/interacted",
                  "provider": provider_id},
                 {"verb": "http://h5p.example.com/expapi/verbs/answered",
                  "provider": provider_id},
             ]
         }, format="json")
        self.assertEqual(response.status_code, 404)

        response = self.provider_client.post(f"/api/v1/provider/analytics-tokens/{token_id}/update-verbs",
         {
             "active_verbs": [
                {"verb": "http://h5p.example.com/expapi/verbs/experienced", "provider": provider_id},
                {"verb": "http://h5p.example.com/expapi/verbs/attempted", "provider": provider_id},
                {"verb": "http://h5p.example.com/expapi/verbs/interacted", "provider": provider_id},
                {"verb": "http://h5p.example.com/expapi/verbs/answered", "provider": provider_id},
            ]
         }, format="json")
        self.assertEqual(response.status_code, 200)

        token = AnalyticsToken.objects.filter(pk=token_id).get()
        self.assertEqual(token.analyticstokenverb_set.count(), 4)

    def test_update_and_delete_token_image(self):
        provider_id = Provider.objects.latest('id').id

        response = self.provider_client.post(
            f"/api/v1/provider/analytics-tokens/create",
            {
                "name": "Test analytics token",
                "description": "Test analytics token",
                "expires": "12/31/2099",
                "can_access": []
            }, format="json"
        )
        self.assertEqual(response.status_code, 201)

        token_id = AnalyticsToken.objects.latest('id').id

        response = self.provider_client.post(f"/api/v1/provider/analytics-tokens/{token_id}/update-verbs",
         {
             "active_verbs": [
                {"verb": "http://h5p.example.com/expapi/verbs/experienced", "provider": provider_id},
                {"verb": "http://h5p.example.com/expapi/verbs/attempted", "provider": provider_id},
                {"verb": "http://h5p.example.com/expapi/verbs/interacted", "provider": provider_id},
                {"verb": "http://h5p.example.com/expapi/verbs/answered", "provider": provider_id},
            ]
         }, format="json")
        self.assertEqual(response.status_code, 200)

        token = AnalyticsToken.objects.filter(pk=token_id).get()
        self.assertEqual(token.analyticstokenverb_set.count(), 4)

        # test upload invalid image
        upload_image = SimpleUploadedFile("image.mp4", b"aslkdjnwekjgfnwkgbwekrjgbwgbkwjebg", content_type="video/mp4")
        response = self.provider_client.post(f"/api/v1/provider/analytics-tokens/{token_id}/image",
                                             {"image": upload_image})
        self.assertEqual(response.status_code, 400)

        # test upload "valid" image
        upload_image = SimpleUploadedFile("image.png", b"aslkdjnwekjgfnwkgbwekrjgbwgbkwjebg", content_type="image/png")
        response = self.provider_client.post(f"/api/v1/provider/analytics-tokens/{token_id}/image",
                                             {"image": upload_image})
        self.assertEqual(response.status_code, 200)

        token = AnalyticsToken.objects.filter(pk=token_id).get()

        # test get image
        response = self.provider_client.get(f"/api/v1/provider/analytics-tokens/{token_id}/image/{token.image_path}")
        self.assertEqual(response.status_code, 200)

        # test get wrong image
        response = self.provider_client.get(f"/api/v1/provider/analytics-tokens/{token_id}/image/nonexistent.png")
        self.assertEqual(response.status_code, 404)

        # test deletion failure for nonexistent token
        response = self.provider_client.delete(f"/api/v1/provider/analytics-tokens/1000000/image")
        self.assertEqual(response.status_code, 404)

        # test deletion
        response = self.provider_client.delete(f"/api/v1/provider/analytics-tokens/{token_id}/image")
        self.assertEqual(response.status_code, 200)

        # test get image
        response = self.provider_client.get(f"/api/v1/provider/analytics-tokens/{token_id}/image/{token.image_path}")
        self.assertEqual(response.status_code, 404)